import compas
from compas.geometry import Primitive

from compas_nurbs.knot_vectors import check_knot_vector
from compas_nurbs.knot_vectors import knot_vector_uniform
from compas_nurbs.knot_vectors import normalize_knot_vector
from compas_nurbs.utilities import prod
from compas_nurbs.utilities import reshape

if not compas.IPY:
    from collections.abc import Iterable

    import numpy as np
    from compas.geometry import transform_points_numpy
else:
    from collections import Iterable


class BSpline(Primitive):
    """A base class for rational and non-rational B-Spline geometry.

    Contains all setters and checkers.
    """

    def __init__(self, control_points, degree, knot_vector, rational, weights=None):
        self.degree = degree                                              # (degree_u, degree_v) for surfaces
        self.__rational = rational
        self.__pdim = len(degree) if isinstance(degree, Iterable) else 1
        self.control_points = control_points                              # 2d for surfaces
        self.knot_vector = knot_vector                                    # (knotvector_u, knotvector_v) for surfaces
        self.weights = weights                                            # 2d for surfaces
        self._build_backend()

    def _build_backend(self):  # needs to be overwritten by derivative classes
        raise NotImplementedError

    @property
    def rational(self):
        return self.__rational

    @property
    def domain(self):
        if self.__pdim == 1:
            return [0., 1.]
        else:
            return [[0., 1.] for _ in range(self.__pdim)]

    @property
    def control_points(self):
        return self._control_points

    @control_points.setter
    def control_points(self, control_points):
        self._control_points = control_points
        if self.__pdim == 1:
            if len(self.control_points) < self.degree + 1:
                raise ValueError("len(control_points) must be >= degree + 1")
        else:
            ok1 = len(self.degree) == len(self.count)
            ok2 = all([c >= d + 1 for c, d in zip(self.count, self.degree)])
            if not all([ok1, ok2]):
                raise ValueError("Invalid control points")

    @property
    def count(self):
        if self.__pdim == 1:
            return len(self.control_points)
        else:
            a, c = self.control_points, []  # lambda?
            for _ in range(self.__pdim):
                c.append(len(a))
                a = a[0]
            return c

    @property
    def knot_vector(self):
        """list of float : The knot vector"""
        return self._knot_vector

    @knot_vector.setter
    def knot_vector(self, knot_vector):
        if self.__pdim == 1:
            if knot_vector:
                if not check_knot_vector(knot_vector, self.count, self.degree):
                    raise ValueError("Invalid knot vector")
                self._knot_vector = normalize_knot_vector(knot_vector)
            else:
                self._knot_vector = knot_vector_uniform(self.count, self.degree)
        else:
            if knot_vector:
                ok1 = all([check_knot_vector(kv, c, d) for kv, c, d in zip(knot_vector, self.count, self.degree)])
                ok2 = len(knot_vector) == len(self.count)
                if not all([ok1, ok2]):
                    raise ValueError("Invalid knot vector")
                self._knot_vector = [normalize_knot_vector(kv) for kv in knot_vector]
            else:
                self._knot_vector = [knot_vector_uniform(c, d) for c, d in zip(self.count, self.degree)]

    @property
    def weights(self):
        return self._weights

    @weights.setter
    def weights(self, weights):
        if self.__pdim == 1:
            if weights:
                if len(weights) != self.count:
                    raise ValueError("Invalid weights! Number of weights must be equal to number of control points.")
                else:
                    self._weights = weights
            else:
                self._weights = [1. for _ in range(self.count)]
        else:
            if not weights:
                self._weights = reshape([1. for _ in range(prod(self.count))], self.count)
            else:
                self._weights = weights  # TODO check

    # ==========================================================================
    # operations
    # ==========================================================================

    def transform(self, transformation):
        xyz = np.array(self.control_points)
        xyz = transform_points_numpy(xyz.reshape(-1, xyz.shape[-1]), transformation)  # TODO more dimen?
        self.control_points = xyz.reshape(self.count + [3])
        self._build_backend()

    def get_bounding_box(self):
        raise NotImplementedError

    def trim(self):
        raise NotImplementedError

    def split(self):
        raise NotImplementedError

    def parameters_at(self):
        raise NotImplementedError

    # ==========================================================================
    # serialisation
    # ==========================================================================

    @property
    def data(self):
        """dict: The data dictionary that represents the bspline geometry."""
        return {'control_points': self.control_points,
                'degree': self.degree,
                'knot_vector': self.knot_vector,
                'rational': self.rational,
                'weights': self.weights}

    @classmethod
    def from_data(cls, data):
        return cls(data['control_points'], data['degree'], data['knot_vector'], data['rational'], data['weights'])



# -*- coding: utf-8 -*-

from sphinx.ext.napoleon.docstring import NumpyDocstring

import sphinx_compas_theme

# -- General configuration ------------------------------------------------

project = 'COMPAS NURBS'
copyright = '2020 - ETH Zurich, Gramazio Kohler Research'
author = 'Romana Rust'
release = '0.1.0'

version = '.'.join(release.split('.')[0:2])

master_doc = 'index'
source_suffix = ['.rst', ]
templates_path = ['_templates']
exclude_patterns = []

pygments_style = 'sphinx'
show_authors = True
add_module_names = True
language = None


# -- Extension configuration ------------------------------------------------

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.mathjax',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    'matplotlib.sphinxext.plot_directive',
]

# autodoc options

autodoc_mock_imports = ["Rhino", "System", "scriptcontext", "rhinoscriptsyntax", "clr", "bpy"]

autodoc_default_options = {
    'undoc-members': True,
    'show-inheritance': True,
}

autodoc_member_order = 'groupwise'

autoclass_content = "class"

# autosummary options

autosummary_generate = True
autosummary_mock_imports = ["Rhino", "System", "scriptcontext", "rhinoscriptsyntax", "clr", "bpy"]

# napoleon options

napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = False
napoleon_use_rtype = False


# first, we define new methods for any new sections and add them to the class
def parse_keys_section(self, section):
    return self._format_fields('Keys', self._consume_fields())


NumpyDocstring._parse_keys_section = parse_keys_section


def parse_attributes_section(self, section):
    return self._format_fields('Attributes', self._consume_fields())


NumpyDocstring._parse_attributes_section = parse_attributes_section


def parse_class_attributes_section(self, section):
    return self._format_fields('Class Attributes', self._consume_fields())


NumpyDocstring._parse_class_attributes_section = parse_class_attributes_section


# we now patch the parse method to guarantee that the the above methods are
# assigned to the _section dict
def patched_parse(self):
    self._sections['keys'] = self._parse_keys_section
    self._sections['class attributes'] = self._parse_class_attributes_section
    self._unpatched_parse()


NumpyDocstring._unpatched_parse = NumpyDocstring._parse
NumpyDocstring._parse = patched_parse

# plot options

# plot_include_source
# plot_pre_code
# plot_basedir
# plot_formats
# plot_rcparams
# plot_apply_rcparams
# plot_working_directory
# plot_template

plot_html_show_source_link = False
plot_html_show_formats = False

# intersphinx options

intersphinx_mapping = {
    'python': ('https://docs.python.org/', None),
    'compas': ('https://compas.dev/compas/latest/', None),
}


# -- Options for HTML output ----------------------------------------------

html_theme = 'compaspkg'
html_theme_path = sphinx_compas_theme.get_html_theme_path()
html_theme_options = {
    "package_name": "compas_nurbs",
    "package_title": project,
    "package_version": release,
    "package_author": "Romana Rust",
    "package_description": "COMPAS package for working with NURBS",
    "package_repo": "https://github.com/gramaziokohler/compas_nurbs",
    "package_docs": "https://gramaziokohler.github.io/compas_nurbs"
}
html_context = {}
html_static_path = []
html_extra_path = ['.nojekyll']
html_last_updated_fmt = ''
html_copy_source = False
html_show_sourcelink = False
html_add_permalinks = ''
html_experimental_html5_writer = True
html_compact_lists = True



from compas_nurbs import Surface
from compas_nurbs.rhino import SurfaceArtist

control_points = [[[0, 0, 0], [0, 4, 0.], [0, 8, -1]],
                  [[3, 0, 6], [3, 4, 0.], [3, 8, 0.]],
                  [[5, 0, 0], [5, 4, 0.], [5, 8, 3.]],
                  [[8, 0, 0], [8, 4, -3], [8, 8, 0.]]]
degree = (3, 2)

surface = Surface(control_points, degree)

artist = SurfaceArtist(surface)
artist.draw()
artist.redraw()



__title__ = 'compas_nurbs'
__description__ = 'NURBS for COMPAS'
__url__ = 'https://github.com/gramaziokohler/compas_nurbs'
__version__ = '0.1.0'
__author__ = 'Gramazio Kohler Research'
__author_email__ = 'gramaziokohler@arch.ethz.ch'
__license__ = 'MIT license'
__copyright__ = 'Copyright 2020 ETH Zurich, Gramazio Kohler Research'

__all__ = ['__author__', '__author_email__', '__copyright__', '__description__', '__license__', '__title__', '__url__', '__version__']



import compas
from compas_nurbs import Curve
from compas_nurbs import RationalCurve
from compas_nurbs import Surface
from compas_nurbs import RationalSurface
from compas_nurbs.utilities import reshape

if compas.RHINO:
    from compas_rhino.geometry import RhinoCurve as BaseRhinoCurve
    from compas_rhino.geometry import RhinoSurface as BaseRhinoSurface

    class RhinoCurve(BaseRhinoCurve):

        @classmethod
        def from_geometry(cls, geometry):
            curve = cls()
            curve.geometry = geometry
            return curve

        def to_compas(self):
            control_points = [[p.X, p.Y, p.Z] for p in self.geometry.Points]
            knot_vector = [k for k in self.geometry.Knots]
            knot_vector = [knot_vector[0]] + knot_vector + [knot_vector[-1]]
            weights = [p.Weight for p in self.geometry.Points]
            if all([w == 1. for w in weights]):
                return Curve(control_points, self.geometry.Degree, knot_vector)
            else:
                return RationalCurve(control_points, self.geometry.Degree, knot_vector, weights=weights)

    class RhinoSurface(BaseRhinoSurface):

        @classmethod
        def from_geometry(cls, geometry):
            surface = cls()
            surface.geometry = geometry
            return surface

        def to_compas(self):
            surface = self.geometry
            knot_vector_u = list(surface.KnotsU)
            knot_vector_v = list(surface.KnotsV)
            knot_vector_u = [knot_vector_u[0]] + knot_vector_u + [knot_vector_u[-1]]
            knot_vector_v = [knot_vector_v[0]] + knot_vector_v + [knot_vector_v[-1]]
            knot_vector = knot_vector_u, knot_vector_v
            degree = surface.Degree(0), surface.Degree(1)
            points = [[p.X, p.Y, p.Z] for p in surface.Points]
            weights = [p.Weight for p in surface.Points]
            count_u = surface.Points.CountU
            count_v = surface.Points.CountV
            points = reshape(points, (count_u, count_v))
            if all([w == 1. for w in weights]):
                return Surface(points, degree, knot_vector)
            else:
                weights = reshape(weights, (count_u, count_v))
                return RationalSurface(points, degree, knot_vector, weights=weights)




import os
from geomdl import BSpline
from geomdl import utilities
from geomdl.visualization import VisMPL
from compas_nurbs import DATA
from compas_nurbs import RationalSurface
from compas_nurbs.evaluators import create_surface


surface = RationalSurface.from_json(os.path.join(DATA, "cylinder.json"))
geomdl_surface = create_surface(surface.control_points, surface.degree, surface.knot_vector, rational=True, weights=surface.weights)


geomdl_surface.delta = 0.001

geomdl_surface.evaluate()
geomdl_surface.sample_size = 40

# Plot the control point polygon and the evaluated curve
vis_comp = VisMPL.VisSurface()
geomdl_surface.vis = vis_comp

# Don't pop up the plot window, instead save it as a PDF file
geomdl_surface.render()



import numpy as np
import rhino3dm
from geomdl import BSpline
from geomdl import NURBS
from compas.geometry import allclose
from compas.geometry import Point
from compas.geometry import Vector
from compas_nurbs import Curve
from compas_nurbs import RationalCurve


def rhino_curve_from_curve(curve):
    rhino_curve = rhino3dm.NurbsCurve(3, curve.rational, curve.degree + 1, curve.count)
    for i, p in enumerate(curve.control_points):
        rhino_curve.Points[i] = rhino3dm.Point4d(*p, curve.weights[i])
    for i, v in enumerate(curve.knot_vector[1:-1]):
        rhino_curve.Knots[i] = v
    return rhino_curve


def geomdl_curve_from_curve(curve):
    crv = NURBS.Curve() if curve.rational else BSpline.Curve()
    crv.degree = curve.degree
    crv.ctrlpts = curve.control_points
    crv.knotvector = curve.knot_vector
    crv.weights = curve.weights
    return crv


def test_curve():
    """Testing Curve against other libraries (geomdl, rhino3dm).
    """
    # settings
    control_points = [[816.32, 139.74, 0.0], [860.41, 238.68, 0.0], [928.17, 199.96, 0.0], [1008.83, 136.51, 0.0],
                      [1062.6, 195.66, 0.0], [1130.35, 248.36, 0.0], [1176.6, 202.11, 0.0], [1240.39, 164.93, 0.0],
                      [1315.67, 242.37, 0.0], [1396.33, 307.97, 0.0], [1466.23, 223.01, 0.0], [1520.01, 156.33, 0.0]]
    degree = 9
    params = np.linspace(0., 1., 10)

    # create curve
    curve = Curve(control_points, degree)
    curve_geomdl = geomdl_curve_from_curve(curve)
    curve_rhino = rhino_curve_from_curve(curve)

    # points
    points = curve.points_at(params)
    rhino_points = [curve_rhino.PointAt(t) for t in params]
    rhino_points = [[p.X, p.Y, p.Z] for p in rhino_points]
    geomdl_points = curve_geomdl.evaluate_list(params)
    assert(allclose(points, rhino_points))
    assert(allclose(points, geomdl_points))

    # derivatives
    geomdl_derivatives = [curve_geomdl.derivatives(u, order=1) for u in params]
    assert(allclose(geomdl_derivatives, curve.derivatives_at(params)))

    # tangents
    tangents = curve.tangents_at(params)
    rhino_tangents = [curve_rhino.TangentAt(t) for t in params]
    rhino_tangents = [[p.X, p.Y, p.Z] for p in rhino_tangents]
    geomdl_tangents = curve_geomdl.tangent(params)
    assert(allclose(tangents, rhino_tangents))
    assert(allclose(tangents, geomdl_tangents))

    # curvature
    curvature = [c.curvature for c in curve.curvatures_at(params)]
    rhino_curvature_vectors = [curve_rhino.CurvatureAt(t) for t in params]
    rhino_curvature = [np.linalg.norm([p.X, p.Y, p.Z]) for p in rhino_curvature_vectors]
    assert(allclose(curvature, rhino_curvature))
    circle_centers = [c.osculating_circle.plane.point for c in curve.curvatures_at(params)]
    rhino_centers = []
    for pt, cv in zip(rhino_points, rhino_curvature_vectors):
        cv = Vector(cv.X, cv.Y, cv.Z)
        rhino_centers.append(Point(*pt) + cv.unitized() * 1/cv.length)
    assert(allclose(circle_centers, rhino_centers))


def test_rational_curve():
    """Testing NurbsCurve against other libraries (geomdl, rhino3dm).

    Unfortunately the rhino3dm Nurbs Curve does not return the correct result.
    Instead rhino_* results for comparison are copied from Rhino directly.
    """
    # settings
    control_points = [[816.32, 139.74, 0.0], [860.41, 238.68, 0.0], [928.17, 199.96, 0.0], [1008.83, 136.51, 0.0],
                      [1062.6, 195.66, 0.0], [1130.35, 248.36, 0.0], [1176.6, 202.11, 0.0], [1240.39, 164.93, 0.0],
                      [1315.67, 242.37, 0.0], [1396.33, 307.97, 0.0], [1466.23, 223.01, 0.0], [1520.01, 156.33, 0.0]]
    degree = 9
    params = np.linspace(0., 1., 10)
    weights = [0.5, 1.1, 0.7, 2., 4., 0.3, 0.1, 0.3, 0.4, 0.5, 1.4, 3.]

    # create curve
    curve = RationalCurve(control_points, degree, weights=weights)
    curve_geomdl = geomdl_curve_from_curve(curve)
    # curve_rhino = rhino_curve_from_curve(curve)

    # points
    points = curve.points_at(params)
    rhino_points = [(816.320, 139.740, 0.), (944.679, 193.328, 0.), (1020.014, 175.383, 0.), (1045.948, 181.534, 0.),
                    (1065.127, 189.409, 0.), (1109.071, 198.051, 0.), (1226.082, 218.523, 0.), (1348.823, 250.108, 0.),
                    (1438.406, 239.541, 0.), (1520.010, 156.330, 0.)]
    geomdl_points = curve_geomdl.evaluate_list(params)
    assert(allclose(points, rhino_points, tol=1e-03))
    assert(allclose(points, geomdl_points))

    # derivatives
    geomdl_derivatives = [curve_geomdl.derivatives(u, order=1) for u in params]
    assert(allclose(geomdl_derivatives, curve.derivatives_at(params)))

    # tangents
    tangents = curve.tangents_at(params)
    rhino_tangents = [(0.407, 0.913, 0.), (0.939, -0.345, 0.), (0.999, 0.033, 0.), (0.917, 0.400, 0.),
                      (0.954, 0.301, 0.), (0.989, 0.151, 0.), (0.977, 0.213, 0.), (0.977, 0.212, 0.),
                      (0.855, -0.518, 0.), (0.628, -0.778, 0.)]
    geomdl_tangents = curve_geomdl.tangent(params)
    assert(allclose(tangents, rhino_tangents, tol=1e-03))
    assert(allclose(tangents, geomdl_tangents))

    # curvature
    curvature = [c.curvature for c in curve.curvatures_at(params)]
    rhino_curvature = [0.000851, 0.000762, 0.013305, 0.006902, 0.009795, 0.00037, 0.000876, 0.003371, 0.008178, 5e-05]
    assert(allclose(curvature, rhino_curvature))
    circle_centers = [c.osculating_circle.plane.point for c in curve.curvatures_at(params)]
    rhino_centers = [(1889.858, -338.654, 0.), (1396.986, 1424.413, 0.), (1017.561, 250.506, 0.), (988.023, 314.332, 0.),
                     (1095.819, 92.043, 0.), (1516.222, -2474.241, 0.), (982.641, 1333.837, 0.), (1411.820, -39.776, 0.),
                     (1375.058, 134.945, 0.), (-14154.272, -12485.585, 0.)]
    assert(allclose(circle_centers, rhino_centers, tol=1e-03))


if __name__ == "__main__":
    test_curve()
    test_rational_curve()



import rhino3dm
from geomdl import BSpline
from geomdl import NURBS
from geomdl import compatibility

from compas.geometry import close
from compas.geometry import allclose
from compas.utilities import flatten

from compas_nurbs import Curve
from compas_nurbs import Surface
from compas_nurbs import RationalSurface
from compas_nurbs.utilities import linspace


def geomdl_surface_from_surface(surface):
    srf = NURBS.Surface() if surface.rational else BSpline.Surface()
    srf.degree_u, srf.degree_v = surface.degree
    if surface.rational:
        control_points = list(flatten(surface.control_points))
        ctrlptsw = compatibility.combine_ctrlpts_weights(control_points, list(flatten(surface.weights)))
        srf.ctrlpts_size_u, srf.ctrlpts_size_v = surface.count
        srf.ctrlptsw = ctrlptsw
    else:
        srf.ctrlpts2d = surface.control_points
    srf.knotvector_u, srf.knotvector_v = surface.knot_vector
    return srf


def rhino_surface_from_surface(surface):
    degree_u, degree_v = surface.degree
    count_u, count_v = surface.count
    knot_vector_u, knot_vector_v = surface.knot_vector
    srf_rhino = rhino3dm.NurbsSurface.Create(3, surface.rational, degree_u + 1, degree_v + 1, count_u, count_v)
    for u, l in enumerate(surface.control_points):
        for v, (x, y, z) in enumerate(l):
            if surface.rational:
                srf_rhino.Points[(u, v)] = rhino3dm.Point4d(x, y, z, surface.weights[u][v])
            else:
                srf_rhino.Points[(u, v)] = rhino3dm.Point4d(x, y, z, 1.)
    for i, v in enumerate(knot_vector_u[1:-1]):
        srf_rhino.KnotsU[i] = v
    for i, v in enumerate(knot_vector_v[1:-1]):
        srf_rhino.KnotsV[i] = v
    return srf_rhino


def test_surface():
    """Testing Surface against other libraries (geomdl, rhino3dm).
    """
    # settings
    control_points_2d = [[[0, 0, 0], [0, 4, 0.], [0, 8, -3]],
                         [[2, 0, 6], [2, 4, 0.], [2, 8, 0.]],
                         [[4, 0, 0], [4, 4, 0.], [4, 8, 3.]],
                         [[6, 0, 0], [6, 4, -3], [6, 8, 0.]]]
    degree_u = 3
    degree_v = 2
    surface = Surface(control_points_2d, (degree_u, degree_v))

    # create surfaces
    srf_geomdl = geomdl_surface_from_surface(surface)
    srf_rhino = rhino_surface_from_surface(surface)

    params_u = linspace(0., 1., 5)
    params_v = linspace(0., 1., 5)
    params = [(u, v) for u in params_u for v in params_v]

    # points
    points = surface.points_at(params)
    geomdl_points = srf_geomdl.evaluate_list(params)
    rhino_points = [srf_rhino.PointAt(*p) for p in params]
    rhino_points = [[p.X, p.Y, p.Z] for p in rhino_points]
    assert(allclose(points, geomdl_points))
    assert(allclose(points, rhino_points))

    # derivatives
    geomdl_derivatives = [srf_geomdl.derivatives(u, v, order=1) for u, v in params]
    assert(allclose(geomdl_derivatives, surface.derivatives_at(params)))

    # normals
    normals = surface.normals_at(params)
    rhino_normals = [srf_rhino.NormalAt(*p) for p in params]
    rhino_normals = [[p.X, p.Y, p.Z] for p in rhino_normals]
    geomdl_normals = [srf_geomdl.normal(p) for p in params]
    assert(allclose(normals, geomdl_normals))
    assert(allclose(normals, rhino_normals))

    # curvature
    curvature = surface.curvatures_at([(0.5, 0.5)])[0]
    rhino_curvature = {'direction': [[-0.93527, 0.3356, 0.11244], [-0.30401, -0.92441, 0.23033]], 'gauss': -
                       0.06905, 'mean': -0.12646, 'kappa': [-0.41808, 0.16516], 'normal': [0.18124, 0.18124, 0.9666]}
    assert(allclose(curvature.direction, rhino_curvature['direction'], tol=0.01))
    assert(allclose(curvature.kappa, rhino_curvature['kappa']))
    assert(allclose(curvature.normal, rhino_curvature['normal'], tol=0.01))
    assert(close(curvature.gauss, rhino_curvature['gauss']))
    assert(close(curvature.mean, rhino_curvature['mean']))


def test_rational_surface():
    """Testing NurbsSurface against other libraries.

    Unfortunately the rhino3dm Nurbs Surface does not return the correct result.
    Instead rhino_* results for comparison are copied from Rhino directly.
    """
    # settings
    control_points_2d = [[[0, 0, 0], [0, 4, 0.], [0, 8, -3]],
                         [[2, 0, 6], [2, 4, 0.], [2, 8, 0.]],
                         [[4, 0, 0], [4, 4, 0.], [4, 8, 3.]],
                         [[6, 0, 0], [6, 4, -3], [6, 8, 0.]]]
    degree = (3, 2)
    weights = [[0.2, 0.1, 0.3], [0.1, 0.7, 1.2], [1.2, 2., 0.4], [0.1, 1.1, 0.5]]
    surface = RationalSurface(control_points_2d, degree, weights=weights)

    # create surfaces
    srf_geomdl = geomdl_surface_from_surface(surface)
    # srf_rhino = rhino_surface_from_surface(surface)  # This surface is wrong

    params_u = linspace(0., 1., 5)
    params_v = linspace(0., 1., 5)
    params = [(u, v) for u in params_u for v in params_v]

    # points
    points = surface.points_at(params)
    geomdl_points = srf_geomdl.evaluate_list(params)
    rhino_points = [[0.0, 0.0, 0.0], [0.0, 1.77778, -0.33333], [0.0, 4.57143, -1.28571], [0.0, 6.85714, -2.31429], [0.0, 8.0, -3.0], [2.58947, 0.0, 0.85263],
                    [2.66072, 2.90039, 0.24467], [2.51172, 4.70621, -0.0269], [2.23804, 6.30153, -0.18823], [1.84305, 8.0, -0.30269], [3.71429, 0.0, 0.42857],
                    [3.67137, 2.70161, -0.00907], [3.52817, 4.19718, -0.07394], [3.25543, 5.67391, 0.05707], [2.67857, 8.0, 0.48214], [4.07735, 0.0, 0.14917],
                    [4.34132, 2.71191, -0.50399], [4.39635, 3.98729, -0.56116], [4.35627, 5.26156, -0.2744], [4.11864, 8.0, 0.88983], [6.0, 0.0, 0.0],
                    [6.0, 3.8, -2.475], [6.0, 4.57143, -2.35714], [6.0, 5.57143, -1.76786], [6.0, 8.0, 0.0]]
    assert(allclose(points, geomdl_points))
    assert(allclose(points, rhino_points))

    # derivatives
    derivatives = surface.derivatives_at(params, order=1)
    geomdl_derivatives = [srf_geomdl.derivatives(u, v, order=1) for u, v in params]
    assert(allclose(geomdl_derivatives, derivatives))

    # normals
    normals = surface.normals_at(params)
    geomdl_normals = [srf_geomdl.normal(p) for p in params]
    rhino_normals = [[-0.94868, 0.0, 0.31623], [-0.66692, 0.19846, 0.71822], [-0.66732, 0.27485, 0.6922], [-0.72062, 0.3195, 0.61533], [-0.76822, 0.38411, 0.51215],
                     [0.15402, 0.21472, 0.96446], [0.09377, 0.17701, 0.97973], [-0.11239, 0.1064, 0.98795], [-0.47128, -0.02486, 0.88163], [-0.77464, -0.16308, 0.61102],
                     [0.56231, 0.15519, 0.81223], [0.45653, 0.11718, 0.88196], [0.28774, 0.01626, 0.95757], [-0.05787, -0.15166, 0.98674], [-0.50649, -0.30352, 0.80706],
                     [0.53366, 0.17691, 0.82699], [0.68907, 0.05998, 0.72221], [0.63712, -0.07255, 0.76734], [0.44518, -0.25925, 0.85709], [0.00122, -0.42503, 0.90518],
                     [-0.0, 0.6, 0.8], [0.74082, 0.11858, 0.66115], [0.81509, -0.23074, 0.5314], [0.78765, -0.34549, 0.51014], [0.76822, -0.38411, 0.51215]]
    assert(allclose(normals, geomdl_normals))
    assert(allclose(geomdl_normals, rhino_normals))

    # curvature
    curvature = surface.curvatures_at([(0.5, 0.5)])[0]
    rhino_curvature = {'direction': [[-0.93542, 0.2192, 0.27737], [-0.20539, -0.97554, 0.07829]], 'gauss': -
                       0.06474, 'mean': -0.14782, 'kappa': [-0.44209, 0.14644], 'normal': [0.28774, 0.01626, 0.95757]}
    assert(allclose(curvature.direction, rhino_curvature['direction'], tol=0.01))
    assert(allclose(curvature.kappa, rhino_curvature['kappa']))
    assert(allclose(curvature.normal, rhino_curvature['normal'], tol=0.01))
    assert(close(curvature.gauss, rhino_curvature['gauss']))
    assert(close(curvature.mean, rhino_curvature['mean']))


def test_loft_surface():
    curves = []
    control_points = [(-24.265, 2.447, 0.000), (-16.799, -7.922, 0.000), (-15.554, 2.613, 0.000), (-2.198, -9.333, 0.000),
                      (-0.456, -2.447, 0.000), (13.646, 4.936, 0.000), (19.370, 0.124, 0.000), (16.550, -8.669, 0.000)]
    curves.append(Curve(control_points, 7))
    control_points = [(-22.606, -7.673, 14.030), (-18.458, 0.456, 14.030), (-8.337, -3.609, 14.030), (-5.434, 6.097, 14.030),
                      (7.673, 0.622, 14.030), (9.167, -6.512, 14.030), (19.868, -3.692, 14.030)]
    curves.append(Curve(control_points, 3))
    control_points = [(-24.845, -9.250, 28.148), (-16.052, 19.536, 28.148), (0.041, -16.799, 28.148), (3.526, 18.209, 28.148),
                      (8.503, 9.167, 28.148), (11.158, 4.687, 28.148), (18.043, -16.882, 28.148), (22.606, -2.696, 28.148)]
    curves.append(Curve(control_points, 7))
    # TODO: loft surface with curves of different degrees and knot vectors
    # surface = Surface.loft_from_curves(curves, degree_v=3)


if __name__ == "__main__":
    test_surface()
    test_rational_surface()
    test_loft_surface()



"""
Main concepts
=============

Curve
-----

.. autoclass:: Curve
   :members:


RationalCurve
-------------

.. autoclass:: RationalCurve
   :members:


Surface
-------

.. autoclass:: Surface
   :members:


RationalSurface
---------------

.. autoclass:: RationalSurface
   :members:


"""

import os

from .__version__ import __version__
from .curve import Curve
from .curve import RationalCurve        # noqa: F401
from .surface import RationalSurface
from .surface import Surface

HERE = os.path.dirname(__file__)
DATA = os.path.join(HERE, "data")

__all_plugins__ = ['compas_nurbs.rhino.install']
__all__ = [
    '__version__',
    'Curve',
    'RationalCurve',
    'RationalSurface',
    'Surface'
]



import compas

from compas.geometry import Point
from compas.geometry import Vector
from compas.geometry import Frame

from compas_nurbs.bspline import BSpline
from compas_nurbs.curvature import CurveCurvature

if not compas.IPY:
    from compas_nurbs.evaluators import create_curve
    from compas_nurbs.evaluators import evaluate_curve
    from compas_nurbs.evaluators import evaluate_curve_derivatives
    from compas_nurbs.operations import curve_tangents
    from compas_nurbs.operations import curve_frames
    from compas_nurbs.operations import curve_curvatures
    from compas_nurbs.fitting import interpolate_curve


class Curve(BSpline):
    """A non-uniform non-rational B-Spline curve.

    Parameters
    ----------
    control_points : list of point
        The curve's control points.
    degree : int
        The degree of the curve.
    knot_vector : list of float, optional
        The knot vector of the curve.

    Attributes
    ----------
    data : dict
        The data representation of the curve.
    control_points : list of class:`compas.geometry.Point`
        The curve's control points.
    degree : int
        The degree of the curve.
    knot_vector : list of float
        The knot vector of the curve.

    Examples
    --------
    >>> control_points = [(0.6, 0.4, 0.0), (0.2, 2.5, 0.0), (6.0, 2.1, 0.0), (4.7, 4.5, 0.0)]
    >>> knot_vector = [0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0]
    >>> degree = 3
    >>> curve = Curve(control_points, degree, knot_vector)
    """

    def __init__(self, control_points, degree, knot_vector=None, rational=False, weights=None):
        super(Curve, self).__init__(control_points, degree, knot_vector, rational, weights)

    def _build_backend(self):
        if not compas.IPY:
            self._curve = create_curve(self.control_points, self.degree, self.knot_vector, self.rational, self.weights)

    # ==========================================================================
    # constructors
    # ==========================================================================

    @classmethod
    def from_points(cls, points, degree, knot_style=0, start_derivative=None, end_derivative=None, periodic=False):
        """Constructs an interpolated curve.

        Parameters
        ----------
        points : list of point
            The list of points on the curve we are looking for.
        degree : int
            The degree of the output parametric curve.
        start_derivative : vector, optional
            The start derivative of the curve. Defaults to ``None``.
        end_derivative : vector
            The end derivative of the curve. Defaults to ``None``.
        knotstyle : int, optional
            The knot style, either 0, 1, or 2 [uniform, chord, or chord_square_root].
            Defaults to 0, uniform.

        Returns
        -------
        :class:`Curve`
            The interpolated curve.

        Examples
        --------
        >>> points = [(0, 0, 0), (3, 4, 0), (-1, 4, 0), (-4, 0, 0), (-4, -3, 0)]
        >>> curve = Curve.from_points(points, 3, knot_style=0)

        """
        cpts, kv = interpolate_curve(points, degree, knot_style, start_derivative, end_derivative, periodic)
        return cls(cpts, degree, kv)

    # ==========================================================================
    # evaluate
    # ==========================================================================

    def points_at(self, params):
        """Evaluates the curve's points at the given parametric positions.

        Parameters
        ----------
        params: list of float
            Evaluation parameters within the curve's domain of [0, 1]

        Returns
        -------
        points : list of :class:`Point`
            Point locations on curve at the parameters.

        Examples
        --------
        >>> curve.points_at([0.0, 0.5, 1.0])
        [Point(0.000, 0.000, 0.000), Point(-0.750, 3.000, 0.000), Point(-4.000, -3.000, 0.000)]
        """
        points = evaluate_curve(self, params)
        return [Point(*p) for p in points]

    def tangents_at(self, params):
        """Evaluates the unit tangent vector at the given parametric positions.

        Parameters
        ----------
        params: list of float

        Returns
        -------
        list of :class:`Vector`
            Unit tangent vectors of the curve at the given parametric positions.

        Examples
        --------
        >>> curve.tangents_at([0.0, 0.5, 1.0])
        [Vector(0.600, 0.800, 0.000), Vector(-0.868, -0.496, 0.000), Vector(0.000, -1.000, 0.000)]
        """
        derivatives = self.derivatives_at(params, order=1)
        tangents = curve_tangents(derivatives)
        return [Vector(*v) for v in tangents]

    def curvatures_at(self, params):
        """Evaluates the curvature at the given parametric positions.

        Parameters
        ----------
        params: list of float

        Returns
        -------
        :class:`CurveCurvature`
            A curvature object with several curvature quantities.

        Examples
        --------
        >>> curvature = curve.curvatures_at([0.5])[0]
        >>> close(curvature.curvature, 0.162835)
        True
        >>> curvature.osculating_circle
        Circle(Plane(Point(2.297, -2.332, 0.000), Vector(0.000, 0.000, 1.000)), 6.141172894211785)
        """
        derivatives = self.derivatives_at(params, order=2)
        curvatures = curve_curvatures(derivatives)
        frames = self.frames_at(params)
        return [CurveCurvature(curvature, frame) for curvature, frame in zip(curvatures, frames)]

    def frames_at(self, params):
        """Evaluates the curve's frames at the given parametric positions.

        Parameters
        ----------
        params: list of float

        Returns
        -------
        list of :class:`Frame`

        Examples
        --------
        >>> curve.frames_at([0.5])
        [Frame(Point(-0.750, 3.000, 0.000), Vector(-0.868, -0.496, 0.000), Vector(0.496, -0.868, 0.000))]
        """
        derivatives = self.derivatives_at(params, order=2)
        points, tangents, normals = curve_frames(derivatives)
        return [Frame(pt, xaxis, yaxis) for pt, xaxis, yaxis in zip(points, tangents, normals)]

    def derivatives_at(self, params, order=1):
        """Evaluates the n-th order curve derivatives at the given parametric positions.

        The output of this method is a list of n-th order derivatives. If order is 0,
        then it will only output the evaluated point. Similarly, if order is 2, then
        it will output the evaluated point, the 1st derivative and the 2nd derivative.

        Parameters
        ----------
        params: list of float
        order: int
            The derivative order.

        Returns
        -------
        :class:`numpy.array`
            A two-dimensional array.

        Examples
        --------
        >>> curve.derivatives_at([0.5], order=0)
        array([[[-0.75,  3.  ,  0.  ]]])
        >>> curve.derivatives_at([0.5], order=1)
        array([[[ -0.75,   3.  ,   0.  ],
                [-10.5 ,  -6.  ,   0.  ]]])
        >>> curve.derivatives_at([0.5], order=2)
        array([[[ -0.75,   3.  ,   0.  ],
                [-10.5 ,  -6.  ,   0.  ],
                [  6.  , -24.  ,   0.  ]]])
        """
        return evaluate_curve_derivatives(self, params, order=order)

    # ==========================================================================
    # operations
    # ==========================================================================

    def reverse(self):
        """Reverses the curve.

        Returns
        -------
        None

        Examples
        --------
        >>> control_points = [(0, 0, 0), (3, 4, 0), (-1, 4, 0), (-4, 0, 0), (-4, -3, 0)]
        >>> curve = Curve(control_points, 3)
        >>> curve.reverse()
        """
        self.control_points = list(reversed(self.control_points))
        self.weights = list(reversed(self.weights))
        self._build_backend()

    # ==========================================================================
    # queries
    # ==========================================================================


class RationalCurve(Curve):
    """A non-uniform rational B-Spline curve (NURBS-Curve).


    Examples
    --------
    >>> control_points = [(0, 0, 0), (3, 4, 0), (-1, 4, 0), (-4, 0, 0), (-4, -3, 0)]
    >>> weights = [0.3, 0.2, 1., 0.4, 2.]
    >>> curve = Curve(control_points, 3, weights=weights)
    """

    def __init__(self, control_points, degree, knot_vector=None, weights=None):
        super(RationalCurve, self).__init__(control_points, degree, knot_vector, rational=True, weights=weights)

    @property
    def weighted_control_points(self):
        """The weighted control points."""
        return [[w * x, w * y, w * z, w] for (x, y, z), w in zip(self.control_points, self.weights)]


if __name__ == '__main__':
    import doctest
    from compas.geometry import allclose  # noqa: F401
    from compas.geometry import close  # noqa: F401
    control_points = [(0, 0, 0), (3, 4, 0), (-1, 4, 0), (-4, 0, 0), (-4, -3, 0)]
    curve = Curve(control_points, 3)
    doctest.testmod(globs=globals())

    control_points = [(0, 0, 0), (3, 4, 0), (-1, 4, 0), (-4, 0, 0), (-4, -3, 0)]
    curve = RationalCurve(control_points, 3)



import scipy
import numpy as np

from .helpers import find_spans
from .helpers import basis_functions
from .helpers import basis_functions_derivatives

from compas.geometry.primitives.curve import binomial_coefficient  # TODO compas: move this upwards

# ==============================================================================
# curve
# ==============================================================================


def create_curve(control_points, degree, knot_vector, rational, weights):
    if not rational:
        return scipy.interpolate.BSpline(knot_vector, control_points, degree)
    else:
        w = np.array([weights]).T
        weighted_control_points = np.concatenate((w * control_points, w), axis=1)
        return scipy.interpolate.BSpline(knot_vector, weighted_control_points, degree)


def evaluate_curve(curve, params):
    """Evaluates a curve at the parameters params.
    """
    points = curve._curve(params)

    if not curve.rational:
        return points
    else:
        w = np.array([points[:, -1]]).T
        return np.delete((1 / w * points), -1, axis=1)


def evaluate_curve_derivatives(curve, params, order=1):
    """Evaluates the n-th order derivatives at the parametric positions `params`.

    Parameters
    ----------
    curve: :class:`scipy.interpolate.BSpline`
        The B-spline curve.
    params : list
        Parametric positions where the derivatives will be computed. Range [0, 1]
    order : int
        The derivative order; to get the i-th derivative

    Returns
    -------
    :class:`numpy.array`
        The evaluated derivatives

    Examples
    --------
    >>>
    """
    derivatives = []
    for i in range(0, order + 1):
        derivatives.append(curve._curve.derivative(i)(params))
    derivatives = np.array(derivatives).transpose(1, 0, 2)

    if not curve.rational:
        return derivatives
    else:
        # TODO: numpify this!
        D = []
        for ders in derivatives:
            new_ders = []
            for k in range(0, order + 1):
                value = ders[k][:-1]
                for i in range(1, k + 1):
                    value[:] = [tmp - (binomial_coefficient(k, i) * ders[i][-1] * drv) for tmp, drv in zip(value, new_ders[k - i])]
                new_ders.append([tmp / ders[0][-1] for tmp in value])
            D.append(new_ders)
        return np.array(D)

# ==============================================================================
# surface
# ==============================================================================


def evaluate_surface(surface, params):
    """Evaluates a surface at the parameters.
    """
    if surface.rational:
        control_points = np.array(surface.weighted_control_points)
    else:
        control_points = np.array(surface.control_points)
    degree_u, degree_v = surface.degree
    knot_vector_u, knot_vector_v = surface.knot_vector
    count_u, count_v = surface.count

    params_u = [p[0] for p in params]
    params_v = [p[1] for p in params]

    spans_u = find_spans(knot_vector_u, count_u, params_u)
    bases_u = basis_functions(degree_u, knot_vector_u, spans_u, params_u)
    spans_v = find_spans(knot_vector_v, count_v, params_v)
    bases_v = basis_functions(degree_v, knot_vector_v, spans_v, params_v)

    points = []
    for span_u, basis_u, span_v, basis_v in zip(spans_u, bases_u, spans_v, bases_v):
        a = basis_v[:degree_v + 1]
        b = control_points[span_u - degree_u:span_u + 1, span_v - degree_v:span_v + 1]
        c = basis_u[:degree_u + 1]
        points.append(np.dot(c, np.dot(a, b)))
    points = np.array(points)

    if not surface.rational:
        return points
    else:
        w = np.array([points[:, -1]]).T
        return np.delete((1 / w * points), -1, axis=1)


def evaluate_surface_derivatives(surface, params, order=1):
    """
    """
    if surface.rational:
        control_points = np.array(surface.weighted_control_points)
    else:
        control_points = np.array(surface.control_points)

    degree_u, degree_v = surface.degree
    knot_vector_u, knot_vector_v = surface.knot_vector
    count_u, count_v = surface.count

    params_u = [p[0] for p in params]
    params_v = [p[1] for p in params]

    spans_u = find_spans(knot_vector_u, count_u, params_u)
    bases_u = basis_functions_derivatives(degree_u, knot_vector_u, spans_u, params_u, order)
    spans_v = find_spans(knot_vector_v, count_v, params_v)
    bases_v = basis_functions_derivatives(degree_v, knot_vector_v, spans_v, params_v, order)

    dv = min(degree_v, order)
    derivatives = []
    for span_u, basis_u, span_v, basis_v in zip(spans_u, bases_u, spans_v, bases_v):
        b = control_points[span_u - degree_u:span_u + 1, span_v - degree_v:span_v + 1]
        temp = np.dot(b.T, np.array(basis_u).T).T
        dd = min(order, dv)
        SKL = np.dot(np.array(basis_v[:dd + 1]), temp[:degree_v + 1]).transpose(1, 0, 2)
        derivatives.append(SKL)

    if not surface.rational:
        return np.array(derivatives)
    else:
        # TODO: numpify this!
        D = []
        for SKLw in derivatives:
            dimension = 4
            SKL = [[[0.0 for _ in range(dimension)] for _ in range(order + 1)] for _ in range(order + 1)]
            for k in range(0, order + 1):
                for l in range(0, order + 1):  # noqa E741
                    v = list(SKLw[k, l])[:]
                    for j in range(1, l + 1):
                        v[:] = [tmp - (binomial_coefficient(l, j) * SKLw[0][j][-1] * drv) for tmp, drv in zip(v, SKL[k][l - j])]
                    for i in range(1, k + 1):
                        v[:] = [tmp - (binomial_coefficient(k, i) * SKLw[i][0][-1] * drv) for tmp, drv in zip(v, SKL[k - i][l])]
                        v2 = [0.0 for _ in range(dimension - 1)]
                        for j in range(1, l + 1):
                            v2[:] = [tmp + (binomial_coefficient(l, j) * SKLw[i][j][-1] * drv) for tmp, drv in zip(v2, SKL[k - i][l - j])]
                        v[:] = [tmp - (binomial_coefficient(k, i) * tmp2) for tmp, tmp2 in zip(v, v2)]
                    SKL[k][l][:] = [tmp / SKLw[0][0][-1] for tmp in v[0:(dimension - 1)]]
            D.append(SKL)
        return np.array(D)


def calculate_surface_curvature(derivatives, order=False):
    """Calculates surface curvature quantities.

    If "order" parameter is set to True, then it will be guaranteed, that C1
    value is always less than C2.

    Note
    ----
    This is copied from https://github.com/nortikin/sverchok
    """
    fu = derivatives[:, 1, 0]
    fv = derivatives[:, 0, 1]
    fuu = derivatives[:, 2, 0]
    fvv = derivatives[:, 0, 2]
    fuv = derivatives[:, 1, 1]

    normal = np.cross(fu, fv)
    norm = np.linalg.norm(normal, axis=1, keepdims=True)
    normal = normal / norm

    nuu = (fuu * normal).sum(axis=1)
    nvv = (fvv * normal).sum(axis=1)
    nuv = (fuv * normal).sum(axis=1)

    duu = np.linalg.norm(fu, axis=1) ** 2
    dvv = np.linalg.norm(fv, axis=1) ** 2
    duv = (fu * fv).sum(axis=1)

    mean = (duu * nvv - 2 * duv * nuv + dvv * nuu) / (2 * (duu * dvv - duv * duv))
    gauss = (nuu * nvv - nuv * nuv) / (duu * dvv - duv * duv)

    n = len(derivatives)  # number of params
    L = np.empty((n, 2, 2))
    L[:, 0, 0] = nuu
    L[:, 0, 1] = nuv
    L[:, 1, 0] = nuv
    L[:, 1, 1] = nvv

    G = np.empty((n, 2, 2))
    G[:, 0, 0] = duu
    G[:, 0, 1] = duv
    G[:, 1, 0] = duv
    G[:, 1, 1] = dvv

    M = np.matmul(np.linalg.inv(G), L)
    eigvals, eigvecs = np.linalg.eig(M)
    # Values of first and second principal curvatures
    c1 = eigvals[:, 0]
    c2 = eigvals[:, 1]

    if order:
        c1mask = (c1 < c2)
        c2mask = np.logical_not(c1mask)
        c1_r = np.where(c1mask, c1, c2)
        c2_r = np.where(c2mask, c1, c2)
    else:
        c1_r = c1
        c2_r = c2

    # dir_1 corresponds to c1, dir_2 corresponds to c2
    dir_1_x = eigvecs[:, 0, 0][np.newaxis].T
    dir_2_x = eigvecs[:, 0, 1][np.newaxis].T
    dir_1_y = eigvecs[:, 1, 0][np.newaxis].T
    dir_2_y = eigvecs[:, 1, 1][np.newaxis].T
    dir_1 = dir_1_x * fu + dir_1_y * fv
    dir_2 = dir_2_x * fu + dir_2_y * fv
    dir_1 = dir_1 / np.linalg.norm(dir_1, axis=1, keepdims=True)
    dir_2 = dir_2 / np.linalg.norm(dir_2, axis=1, keepdims=True)

    if order:
        c1maskT = c1mask[np.newaxis].T
        c2maskT = c2mask[np.newaxis].T
        dir_1_r = np.where(c1maskT, dir_1, -dir_2)
        dir_2_r = np.where(c2maskT, dir_1, dir_2)
    else:
        dir_1_r = dir_1
        dir_2_r = dir_2

    return c1_r, c2_r, dir_1_r, dir_2_r, normal, mean, gauss



def find_span(knot_vector, number_of_control_points, t):
    """Finds the span of a single knot over the knot vector using linear search.

    Alternative implementation for the Algorithm A2.1 from The NURBS Book by Piegl & Tiller.

    # TODO find_span_binsearch
    """
    span = 0
    while span < number_of_control_points and knot_vector[span] <= t:
        span += 1
    return span - 1


def find_spans(knot_vector, number_of_control_points, params):
    """
    TODO: check if np.vectorize makes it faster.
    """
    return list(map(lambda t: find_span(knot_vector, number_of_control_points, t), params))


def basis_function(degree, knot_vector, span, t):
    """Computes the non-vanishing basis functions for a single parameter t.

    Implementation of Algorithm A2.2 from The NURBS Book by Piegl & Tiller.
    Uses recurrence to compute the basis functions, also known as Cox - de
    Boor recursion formula.
    """
    left = [0.0 for _ in range(degree + 1)]
    right = [0.0 for _ in range(degree + 1)]
    N = [1.0 for _ in range(degree + 1)]

    for j in range(1, degree + 1):
        left[j] = t - knot_vector[span + 1 - j]
        right[j] = knot_vector[span + j] - t
        saved = 0.0
        for r in range(0, j):
            temp = N[r] / (right[r + 1] + left[j - r])
            N[r] = saved + right[r + 1] * temp
            saved = left[j - r] * temp
        N[j] = saved
    return N


def basis_function_derivatives(degree, knot_vector, span, t, order):
    """Computes derivatives of the basis functions for a single parameter t.

    Implementation of Algorithm A2.3 from The NURBS Book by Piegl & Tiller.

    Parameters
    ----------
    degree: int
    knot_vector: knot vector
    span: int
    t: float
    order: int
        The order of the derivative.

    Returns
    -------
    list : derivatives of the basis functions
    """
    # Initialize variables
    left = [1.0 for _ in range(degree + 1)]
    right = [1.0 for _ in range(degree + 1)]
    ndu = [[1.0 for _ in range(degree + 1)] for _ in range(degree + 1)]  # N[0][0] = 1.0 by definition

    for j in range(1, degree + 1):
        left[j] = t - knot_vector[span + 1 - j]
        right[j] = knot_vector[span + j] - t
        saved = 0.0
        r = 0
        for r in range(r, j):
            # Lower triangle
            ndu[j][r] = right[r + 1] + left[j - r]
            temp = ndu[r][j - 1] / ndu[j][r]
            # Upper triangle
            ndu[r][j] = saved + (right[r + 1] * temp)
            saved = left[j - r] * temp
        ndu[j][j] = saved

    # Load the basis functions
    ders = [[0.0 for _ in range(degree + 1)] for _ in range((min(degree, order) + 1))]
    for j in range(0, degree + 1):
        ders[0][j] = ndu[j][degree]

    # Start calculating derivatives
    a = [[1.0 for _ in range(degree + 1)] for _ in range(2)]
    # Loop over function index
    for r in range(0, degree + 1):
        # Alternate rows in array a
        s1 = 0
        s2 = 1
        a[0][0] = 1.0
        # Loop to compute k-th derivative
        for k in range(1, order + 1):
            d = 0.0
            rk = r - k
            pk = degree - k
            if r >= k:
                a[s2][0] = a[s1][0] / ndu[pk + 1][rk]
                d = a[s2][0] * ndu[rk][pk]
            if rk >= -1:
                j1 = 1
            else:
                j1 = -rk
            if (r - 1) <= pk:
                j2 = k - 1
            else:
                j2 = degree - r
            for j in range(j1, j2 + 1):
                a[s2][j] = (a[s1][j] - a[s1][j - 1]) / ndu[pk + 1][rk + j]
                d += (a[s2][j] * ndu[rk + j][pk])
            if r <= pk:
                a[s2][k] = -a[s1][k - 1] / ndu[pk + 1][r]
                d += (a[s2][k] * ndu[r][pk])
            ders[k][r] = d

            # Switch rows
            j = s1
            s1 = s2
            s2 = j

    r = float(degree)  # Multiply through by the the correct factors
    for k in range(1, order + 1):
        for j in range(0, degree + 1):
            ders[k][j] *= r
        r *= (degree - k)
    return ders


def basis_functions(degree, knot_vector, spans, params):
    """
    TODO: check if np.vectorize makes it faster.
    """
    return list(map(lambda span, t: basis_function(degree, knot_vector, span, t), spans, params))


def basis_functions_derivatives(degree, knot_vector, spans, params, order):
    """
    TODO: check if np.vectorize makes it faster.
    """
    return list(map(lambda span, t: basis_function_derivatives(degree, knot_vector, span, t, order), spans, params))


"""

or base on scipy's bspline?
https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.BSpline.html

BSline.basis_element(t[, extrapolate])activa
BSline.derivative(self[, nu])
BSline.integrate(self, a, b[, extrapolate])

"""



import time

import Rhino.Geometry as rg
import rhinoscriptsyntax as rs
import scriptcontext as sc

from compas_rhino.artists import BaseArtist


class CurveArtist(BaseArtist):
    """
    """

    def __init__(self, curve):
        self.curve = curve

    def draw(self):
        points = [rg.Point3d(*p) for p in self.curve.control_points]
        knots = self.curve.knot_vector[1:-1]
        return rs.AddNurbsCurve(points, knots, self.curve.degree, self.curve.weights)

    def draw_points(self):
        return rs.AddPoints(self.curve.control_points)

    def redraw(self, timeout=None):
        """Redraw the Rhino view.

        Parameters
        ----------
        timeout : float, optional
            The amount of time the artist waits before updating the Rhino view.
            The time should be specified in seconds.
            Default is ``None``.

        """
        if timeout:
            time.sleep(timeout)

        sc.doc.Views.Redraw()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import compas.plugins


@compas.plugins.plugin(category='install')
def installable_rhino_packages():
    return ['compas_nurbs']



import pytest
from compas.geometry import close
from compas.geometry import allclose
from compas_nurbs import Surface
from compas_nurbs import Curve


@pytest.fixture(autouse=True)
def add_imports(doctest_namespace):
    doctest_namespace["Surface"] = Surface
    doctest_namespace["Curve"] = Curve
    doctest_namespace["close"] = close
    doctest_namespace["allclose"] = allclose


@pytest.fixture(scope='function', autouse=True)
def create_nurbs(request, doctest_namespace):
    if request.module.__name__ in ('compas_nurbs.surface', 'compas_nurbs.curve'):
        control_points = [(0, 0, 0), (3, 4, 0), (-1, 4, 0), (-4, 0, 0), (-4, -3, 0)]
        curve = Curve(control_points, 3)
        doctest_namespace["curve"] = curve

        control_points_2d = [[[0, 0, 0], [0, 4, 0.], [0, 8, -3]],
                             [[2, 0, 6], [2., 4, 0.], [2, 8, 0.]],
                             [[4, 0, 0], [4., 4, 0.], [4, 8, 3.]],
                             [[6, 0, 0], [6., 4, -3], [6, 8, 0.]]]
        degree_u, degree_v = 3, 2
        surface = Surface(control_points_2d, (degree_u, degree_v))
        doctest_namespace["surface"] = surface
        yield
    else:
        yield



import time

import Rhino.Geometry as rg
import rhinoscriptsyntax as rs
import scriptcontext as sc

from compas.utilities import flatten
from compas_rhino.artists import BaseArtist


class SurfaceArtist(BaseArtist):
    """
    """

    def __init__(self, surface):
        self.surface = surface

    def draw(self):
        point_count = self.surface.count
        points = [rg.Point3d(*p) for pl in self.surface.control_points for p in pl]
        weights = list(flatten(self.surface.weights))
        knots_u = self.surface.knot_vector[0][1:-1]
        knots_v = self.surface.knot_vector[1][1:-1]
        degree = self.surface.degree
        ns = rg.NurbsSurface.Create(3, weights is not None, degree[0] + 1, degree[1] + 1, point_count[0], point_count[1])
        index = 0
        for i in range(point_count[0]):
            for j in range(point_count[1]):
                if weights:
                    cp = rg.ControlPoint(points[index], weights[index])
                    ns.Points.SetControlPoint(i, j, cp)
                else:
                    cp = rg.ControlPoint(points[index])
                    ns.Points.SetControlPoint(i, j, cp)
                index += 1
        for i in range(ns.KnotsU.Count):
            ns.KnotsU[i] = knots_u[i]
        for i in range(ns.KnotsV.Count):
            ns.KnotsV[i] = knots_v[i]
        assert(ns.IsValid)

        sc.doc.Objects.AddSurface(ns)

    def draw_points(self):
        return rs.AddPoints(self.surface.control_points)

    def redraw(self, timeout=None):
        """Redraw the Rhino view.

        Parameters
        ----------
        timeout : float, optional
            The amount of time the artist waits before updating the Rhino view.
            The time should be specified in seconds.
            Default is ``None``.

        """
        if timeout:
            time.sleep(timeout)

        sc.doc.Views.Redraw()



import numpy as np


def normalize_vectors(vectors):
    return (vectors.T / np.linalg.norm(vectors, axis=1).reshape(1, -1)).T


def curve_tangents(derivatives):
    return normalize_vectors(derivatives[:, 1])


def curve_frames(derivatives):
    points, d1, d2 = derivatives[:, 0], derivatives[:, 1], derivatives[:, 2]
    binormal = normalize_vectors(np.cross(d1, d2, axis=1))
    tangents = normalize_vectors(np.array(d1))
    normals = np.cross(binormal, tangents, axis=1)
    return points, tangents, normals


def curve_curvatures(derivatives):
    d1, d2 = derivatives[:, 1], derivatives[:, 2]
    return np.linalg.norm(np.cross(d1, d2, axis=1), axis=1) / np.linalg.norm(d1, axis=1)**3


def curve_is_planar(curve):
    """Returns ``True`` if the curve is planar.
    """
    # TODO: only checks if the curve is planar in xy-, xz-, yz-plane, oriented bounding box to check?
    ranges = np.ptp(curve.control_points, axis=0)
    is_planar = not ranges.all()
    ix, iy, iz = np.argsort(ranges)[::-1]
    return is_planar, np.take(curve.control_points, [ix, iy], axis=1)


def unify_curves(curves):
    """
    """
    # 1. Raise degree to maximum degree
    max_degree = max([crv.degree for crv in curves])
    knot_vector = curves[0].knot_vector
    for crv in curves:
        if crv.degree != max_degree:
            raise NotImplementedError("Sorry, lofting curves of different degrees is currently not yet implemented")
        if crv.knot_vector != knot_vector:
            raise NotImplementedError("Sorry, lofting curves with different knot vectors is currently not yet implemented")
    # curve_elevate_degree(curve, final_degree)
    # unify knot vector
    return curves


def curve_elevate_degree(curve, final_degree):
    raise NotImplementedError


def surface_normals(surface, params):
    skl = surface.derivatives_at(params, order=1)
    vectors = np.cross(skl[:, 1, 0], skl[:, 0, 1])
    return normalize_vectors(vectors)



import math
from compas.geometry import Bezier
from compas.geometry import distance_point_point


class CurveKnotStyle(object):
    """
    """
    Uniform = 0
    Chord = 1
    ChordSquareRoot = 2

    def __init__(self, style, periodic=False):
        self.style = style
        self.periodic = periodic

    # TODO add methods ?


def equally_spaced_parameters(num_points):
    """Computes equally spaced parameters for interpolation.

    Please refer to the equation 9.3 on The NURBS Book (2nd Edition), pp.364.

    Parameters
    ----------
    num_points : int
        Number of points to compute parameters for.

    Returns
    -------
    list of float
        Parameters within the range of [0, 1].
    """
    n = num_points - 1
    return [i / float(n) for i in range(n + 1)]


def bezier_spaced_parameters(num_points):
    """Computes bezier spaced parameters for interpolation.

    Parameters
    ----------
    num_points : int
        Number of points to compute parameters for.

    Returns
    -------
    list of float
        Parameters within the range of [0, 1].

    Notes
    -----
    If degree >= 5 this provides a more stable knotvector for 'uniform' knot style.
    """
    curve = Bezier([[0.0, 0.0, 0.0], [0.5, 0.0, 0.0], [0.5, 1.0, 0.0], [1., 1., 0]])
    return [curve.point(t).y for t in equally_spaced_parameters(num_points)]


def chord_lengths(points):
    """Returns the chord lengths.

    Parameters
    ----------
    points : list of point
        Points on the curve.

    Returns
    -------
    list of float
        The n - 1 lengths between the n points.

    Examples
    --------
    >>>
    """
    return [distance_point_point(a, b) for a, b in zip(points[:-1], points[1:])]


def chord_spaced_parameters(points):
    """Computes parameters based on chord length for interpolation.

    Please refer to the Equations 9.4 and 9.5 for chord length parametrization,
    and Equation 9.6 for centripetal method on The NURBS Book (2nd Edition),
    pp.364-365.

    Parameters
    ----------
    points : list of point
        Points on the curve.

    Returns
    -------
    list of float
        Parameters within the range of [0, 1].

    Examples
    --------
    >>>
    """
    cds = chord_lengths(points)
    d = sum(cds)
    uk = [0.] + cds + [1.]
    return [sum(uk[0:i + 1]) / d for i in range(len(points))]


def chord_sqrt_spaced_parameters(points):
    """Computes parameters based on the chord length for interpolation.

    Please refer to the Equations 9.4 and 9.5 for chord length parametrization,
    and Equation 9.6 for centripetal method on The NURBS Book (2nd Edition),
    pp.364-365.

    Parameters
    ----------
    points : list of point
        Points on the curve.

    Returns
    -------
    list of float
        Parameters within the range of [0, 1].

    Examples
    --------
    >>>
    """
    cds = [math.sqrt(v) for v in chord_lengths(points)]
    d = sum(cds)
    uk = [0.] + cds + [1.]
    return [sum(uk[0:i + 1]) / d for i in range(len(points))]


def knot_vector_uniform(num_points, degree, periodic=False):
    """Computes a uniform knot vector.

    Parameters
    ----------
    num_points : int
        Number of points to compute parameters for.
    degree : int
        The degree of the curve.

    Returns
    -------
    list of float
        The knot vector in the domain of [0, 1].

    Notes
    -----
    Is the same to which Rhino refers to as CurveKnotStyle.Uniform

    """
    kv = [0.0 for _ in range(degree + 1)]
    step = 1. / (num_points - degree)
    for i in range(1, num_points - degree):
        kv.append(step * i)
    kv += [1.0 for _ in range(degree + 1)]
    return kv


def knot_vector_from_params(degree, params, periodic=False):
    """Computes a knot vector from parameters using the averaging method.

    Please refer to the Equation 9.8 on The NURBS Book (2nd Edition), pp.365 for
    details.

    Parameters
    ----------
    degree : int
        The degree of the curve
    params : list of float
        Parameters on the curve in the range of [0, 1].

    Returns
    -------
    list of float
        The knot vector.

    Notes
    -----
    Is the same as geomdl.fitting.compute_knot_vector
    """
    kv = [0.0 for _ in range(degree + 1)]
    for j in range(1, len(params) - degree):
        v = (1.0 / degree) * sum([params[j] for j in range(j, j + degree)])
        kv.append(v)
    kv += [1.0 for _ in range(degree + 1)]
    return kv


def knot_vector_and_params(points, degree, knot_style, extended=False, periodic=False):
    """Returns the knot vector for curve interpolation.

    Parameters
    ----------
    points : list of point
        Points on the curve.
    degree : int
        The degree of the curve
    knot_style : int
        The knot style: 0 for 'Uniform', 1 for 'Chord', 2 for 'ChordSquareRoot'
    extended : bool
        `True` if the knot vector and params should be calculated for interpolation
        with end derivatives. Defaults to `False`.
    periodic : bool
        `True` if knot vector should be computed for a `closed` curve. Defaults to
        `False`.

    Returns
    -------
    tuple : (knot_vector, parameters)
        The knot vector and the parameters for the interpolation.
    """

    if knot_style == CurveKnotStyle.Uniform:
        uk = equally_spaced_parameters(len(points))
    elif knot_style == CurveKnotStyle.Chord:
        uk = chord_spaced_parameters(points)
    elif knot_style == CurveKnotStyle.ChordSquareRoot:
        uk = chord_sqrt_spaced_parameters(points)
    else:
        raise ValueError("CurveKnotStyle %d is unknown" % knot_style)

    if extended:  # extend parameters for end derivatives estimation
        uk = [uk[0], 0.] + uk[1:-1] + [1., uk[-1]]

    # if knot_style == CurveKnotStyle.Uniform:
    #    kv = knot_vector_uniform(len(uk), degree)
    # else:
    kv = knot_vector_from_params(degree, uk)

    return kv, uk


def normalize_knot_vector(knot_vector):
    """Returns a normalized knot vector within the [0, 1] domain.

    Parameters
    ----------
    list of float
        A knot vector

    Returns
    -------
    list of float
        The normalized knot vector.
    """
    knot_vector = [v - knot_vector[0] for v in knot_vector]
    return [v / float(knot_vector[-1]) for v in knot_vector]


def check_knot_vector(knot_vector, num_points, degree):
    """Checks the validity of the knot vector.

    Parameters
    ----------
    list of float
        The knot vector to check.
    num_points : int
        Number of points to compute parameters for.
    degree : int
        The degree

    Returns
    -------
    bool
        True if the knot vector is valid, False otherwise.
    """
    if len(knot_vector) != degree + num_points + 1:  # m = p + n + 1
        return False

    # Check ascending order
    prev_knot = knot_vector[0]
    for knot in knot_vector:
        if prev_knot > knot:
            return False
        prev_knot = knot
    return True




import compas
from compas.geometry import Shape, Vector, Point

from compas_nurbs.bspline import BSpline
from compas_nurbs.curve import Curve
from compas_nurbs.curvature import SurfaceCurvature

if not compas.IPY:
    from compas_nurbs.evaluators import evaluate_surface
    from compas_nurbs.evaluators import evaluate_surface_derivatives
    from compas_nurbs.evaluators import calculate_surface_curvature
    from compas_nurbs.operations import surface_normals
    from compas_nurbs.operations import unify_curves


class Surface(BSpline, Shape):
    """A base class for n-variate B-spline (non-rational) surfaces.

    Parameters
    ----------
    control_points : list of list of point
        The surface' 2-dimensional array of control points.
    degree : tuple of int
        Surface degree for the u- and the v-direction: (u, v).
    knot_vector : tuple of list of float, optional
        Surface knot vectors. One for the u- and one for the v-direction: (ku, kv).

    Attributes
    ----------
    data : dict
        The data representation of the surface.
    control_points : list of list of point
        The surface' 2-dimensional array of control points.
    degree : tuple of int
        Surface degree for the u- and the v-direction: (u, v).
    knot_vector : tuple of list of float
        Surface knot vectors. One for the u- and one for the v-direction: (ku, kv).

    Examples
    --------
    >>> control_points_2d = [[[0, 0, 0], [0, 4, 0], [0, 8, -3]], [[2, 0, 6], [2, 4, 0], [2, 8, 0]], [[4, 0, 0], [4, 4, 0], [4, 8, 3]], [[6, 0, 0], [6, 4, -3], [6, 8, 0]]]
    >>> degree_u, degree_v = 3, 2
    >>> surface = Surface(control_points_2d, (degree_u, degree_v))
    """

    def __init__(self, control_points, degree, knot_vector=None, rational=False, weights=None):
        super(Surface, self).__init__(control_points, degree, knot_vector, rational, weights)

    def _build_backend(self):
        self._surface = self  # no numpy surface

    # ==========================================================================
    # constructors
    # ==========================================================================

    @classmethod
    def loft_from_curves(cls, curves, degree_v=3):
        """Creates a :class:`Surface` by lofting between curves.

        Parameters
        ----------
        curves: list of :class:`Curve`
            A list of curves.
        degree_v : int
            The degree of the resulting surface in v-direction.

        Returns
        -------
        :class:`Surface`
            The resulting surface.

        Examples
        --------
        >>>
        """
        curves = unify_curves(curves)
        degree_u = curves[0].degree
        knot_vector_u = curves[0].knot_vector
        degree_v = min(degree_v, len(curves) - 1)
        control_points = []
        count_u = len(curves[0].control_points)
        knot_vector_v = []
        for i in range(count_u):
            points = [crv.control_points[i] for crv in curves]
            c = Curve.from_points(points, degree_v)
            control_points.append(c.control_points)
            knot_vector_v = knot_vector_v or c.knot_vector
        # Rhino lofts into the opposite direction (u=>v)
        return cls(control_points, (degree_u, degree_v), (knot_vector_u, knot_vector_v))

    @classmethod
    def from_points(cls, points):
        pass

    # ==========================================================================
    # evaluate
    # ==========================================================================

    def points_at(self, params):
        """Evaluates the surface's points at the given parametric positions.

        Parameters
        ----------
        params: list of tuples (u, v)
            Evaluation parameters within the curve's domain of [0, 1]

        Returns
        -------
        points : list of :class:`Point`
            Point locations on the surface at the parameters.

        Examples
        --------
        >>> params = [(0.1, 0.1), (0.1, 0.5), (0.5, 0.1), (0.5, 0.5)]
        >>> surface.points_at(params)
        [Point(0.600, 0.800, 1.159), Point(0.600, 4.000, -0.164), Point(3.000, 0.800, 1.763), Point(3.000, 4.000, 0.562)]
        """
        points = evaluate_surface(self._surface, params)
        return [Point(*p) for p in points]

    def normals_at(self, params):
        """Evaluates the surface's normals at the given parametric positions.

        Parameters
        ----------
        params: list of tuples (u, v)
            Evaluation parameters within the curve's domain of [0, 1]

        Returns
        -------
        vector : list of :class:`Vector`
            Normalized surface normals at the parameters.

        Examples
        --------
        >>> params = [(0.1, 0.1), (0.1, 0.5), (0.5, 0.1), (0.5, 0.5)]
        >>> surface.normals_at(params)
        [Vector(-0.822, 0.203, 0.533), Vector(-0.605, 0.324, 0.727), Vector(0.503, 0.424, 0.753), Vector(0.181, 0.181, 0.967)]
        """
        normals = surface_normals(self._surface, params)
        return [Vector(*n) for n in normals]

    def curvatures_at(self, params):
        """Evaluates the surface' curvature at the given parametric positions.

        Parameters
        ----------
        params: list of tuples (u, v)
            Evaluation parameters within the curve's domain of [0, 1]

        Returns
        -------
        :class:`SurfaceCurvature`
            A curvature object with several curvature quantities.

        Examples
        --------
        >>> params = [(0.5, 0.5)]
        >>> curvature = surface.curvatures_at(params)[0]
        >>> curvature.direction # principal curvature directions
        [Vector(-0.935, 0.336, 0.112), Vector(-0.304, -0.924, 0.230)]
        >>> allclose(curvature.kappa, [-0.41808, 0.16516]) # principal curvature values
        True
        >>> curvature.normal
        Vector(0.181, 0.181, 0.967)
        >>> close(curvature.gauss, -0.06905)
        True
        >>> close(curvature.mean, -0.12646)
        True
        """
        derivatives = self._surface.derivatives_at(params, order=2)
        kappa1, kappa2, direction1, direction2, normal, mean, gauss = calculate_surface_curvature(derivatives)
        return [SurfaceCurvature((k1, k2), (d1, d2), n, m, g) for k1, k2, d1, d2, n, m, g in zip(kappa1, kappa2, direction1, direction2, normal, mean, gauss)]

    def derivatives_at(self, params, order=1):
        """Evaluates n-th order surface derivatives at the given (u, v) parameter pairs.

        Parameters
        ----------
        params : list of (u, v) tuples
            The parameters to evaluate in the [0, 1] domain.
        order : int
            The derivative order.

        Returns
        -------
        :class:`numpy.array`
            A two-dimensional array.
        """
        return evaluate_surface_derivatives(self._surface, params, order=order)

    # ==========================================================================
    # operations
    # ==========================================================================

    # ==========================================================================
    # queries
    # ==========================================================================

    # ==========================================================================
    # serialisation
    # ==========================================================================


class RationalSurface(Surface):

    def __init__(self, control_points, degree, knot_vector=None, rational=True, weights=None):
        super(RationalSurface, self).__init__(control_points, degree, knot_vector, True, weights)

    @property
    def weighted_control_points(self):
        return [[[w * x, w * y, w * z, w] for (x, y, z), w in zip(cl, wl)] for cl, wl in zip(self.control_points, self.weights)]


if __name__ == "__main__":

    import doctest
    from compas.geometry import allclose  # noqa: F401
    from compas.geometry import close  # noqa: F401
    control_points = [[[0, 0, 0], [0., 4, 0.], [0, 8, -3]],
                      [[2, 0, 6], [2., 4, 0.], [2, 8, 0.]],
                      [[4, 0, 0], [4., 4, 0.], [4, 8, 3.]],
                      [[6, 0, 0], [6., 4, -3], [6, 8, 0.]]]
    degree = (3, 2)
    surface = Surface(control_points, degree)
    params = [(u, v) for u in [0, 0.5, 1] for v in [0, 0.5, 1]]

    doctest.testmod(globs=globals())



import numpy as np
from scipy.linalg import lu_factor, lu_solve

from compas.geometry import allclose  # noqa: F401

from compas_nurbs.helpers import find_spans
from compas_nurbs.helpers import basis_functions
from compas_nurbs.knot_vectors import knot_vector_and_params
from compas_nurbs.knot_vectors import CurveKnotStyle


# TODO: estimate derivatives for degree==3
# https://link.springer.com/content/pdf/10.1007/s003660050038.pdf

def global_curve_interpolation(points, degree, knot_style=0, periodic=False):
    """Global curve interpolation through points.

    Please refer to Algorithm A9.1 on The NURBS Book (2nd Edition), pp.369-370
    for details.

    Parameters
    ----------
    points : list of point
        The list of points on the curve we are looking for.
    degree : int
        The degree of the output parametric curve.
    knotstyle : int, optional
        The knot style, either 0, 1, or 2 [uniform, chord, or chord_square_root].
        Defaults to 0, uniform.

    Returns
    -------
    tuple (control_points, knot_vector)

    Examples
    --------
    >>> degree = 3
    >>> points = [(0., 0., 0.), (3., 4., 0.), (-1., 4., 0.), (-4., 0., 0.), (-4., -3., 0.)]
    >>> control_points, knot_vector = global_curve_interpolation(points, degree)
    >>> solution = [(0, 0, 0), (6.44, 3.72, 0.), (-2.67, 7.5, 0), (-5.11, -2.72, 0), (-4, -3, 0)]
    >>> allclose(control_points, solution, tol=0.005)
    True
    """
    points = np.array(points)
    kv, uk = knot_vector_and_params(points, degree, knot_style, extended=False)
    M = coefficient_matrix(degree, kv, uk)
    lu, piv = lu_factor(M)
    return lu_solve((lu, piv), points), kv


def global_curve_interpolation_with_end_derivatives(points,
                                                    degree,
                                                    start_derivative,
                                                    end_derivative,
                                                    knot_style=0,
                                                    periodic=False):
    """Global curve interpolation through points with end derivatives specified.

    Please refer to The NURBS Book (2nd Edition), pp. 370-371 for details.

    Parameters
    ----------
    points : list of point
        The list of points on the curve we are looking for.
    degree : int
        The degree of the output parametric curve.
    start_derivative : vector
        The start derivative of the curve
    end_derivative : vector
        The end derivative of the curve
    knotstyle : int, optional
        The knot style, either 0, 1, or 2 [uniform, chord, or chord_square_root].
        Defaults to 0, uniform.

    Returns
    -------
    tuple (control_points, knot_vector)

    Examples
    --------
    >>> degree = 3
    >>> points = [(0., 0., 0.), (3., 4., 0.), (-1., 4., 0.), (-4., 0., 0.), (-4., -3., 0.)]
    >>> start_derivative = [17.75, 10.79, 0.0]
    >>> end_derivative = [-0.71, -12.62, 0.0]
    >>> result = global_curve_interpolation_with_end_derivatives(points, degree, start_derivative, end_derivative)
    >>> solution = [(0, 0, 0), (1.48, 0.9, 0), (4.95, 5.08, 0), (-1.56, 4.87, 0), (-4.72, -0.56, 0), (-3.94, -1.95, 0), (-4, -3, 0)]
    >>> allclose(result[0], solution, tol=0.005)
    True
    """

    points = np.array(points)
    kv, uk = knot_vector_and_params(points, degree, knot_style, extended=True)
    M = coefficient_matrix(degree, kv, uk)

    M[1][0] = -1.
    M[1][1] = 1.
    M[-2][-2] = -1.
    M[-2][-1] = 1.

    v0 = np.array(start_derivative) * kv[degree + 1] / degree
    vn = np.array(end_derivative) * (1 - kv[len(kv) - 1 - degree - 1]) / degree
    C = points[:]
    C = np.insert(C, 1, v0, axis=0)
    C = np.insert(C, -1, vn, axis=0)

    lu, piv = lu_factor(M)
    return lu_solve((lu, piv), C), kv


def coefficient_matrix(degree, knot_vector, uk):
    """Returns the coefficient matrix for global interpolation.

    Please refer to The NURBS Book (2nd Edition), pp. 370-371 for details.

    Parameters
    ----------
    degree : int
        The degree of the curve.
    knot_vector : list of float
        The knot vector of the curve.
    uk : list of float
        parameters

    Returns
    -------
    coefficient matrix, list
    """
    # TODO: use this for evaluators?
    num_points = len(uk)
    spans = find_spans(knot_vector, num_points, uk)
    bases = basis_functions(degree, knot_vector, spans, uk)
    M = [[0.0 for _ in range(num_points)] for _ in range(num_points)]
    for i, (span, basis) in enumerate(zip(spans, bases)):
        M[i][span - degree:span + 1] = basis[:degree + 1]
    return np.array(M)


def interpolate_curve(points, degree, knot_style=0, start_derivative=None, end_derivative=None, periodic=False):
    """Interpolate curve by the specified parameters.

    Parameters
    ----------
    points : list of point
        The list of points on the curve we are looking for.
    degree : int
        The degree of the output parametric curve.
    start_derivative : vector, optional
        The start derivative of the curve. Defaults to ``None``.
    end_derivative : vector
        The end derivative of the curve. Defaults to ``None``.
    knotstyle : int, optional
        The knot style, either 0, 1, or 2 [uniform, chord, or chord_square_root].
        Defaults to 0, uniform.

    Returns
    -------
    tuple (control_points, knot_vector)


    Raises
    ------
    ValueError
        If the knot style is not correct or if only one derivative is passed.
    """
    if knot_style not in [CurveKnotStyle.Uniform, CurveKnotStyle.Chord, CurveKnotStyle.ChordSquareRoot]:
        raise ValueError("Please pass a valid knot style: [0, 1, 2].")
    if start_derivative is not None and end_derivative is not None:
        return global_curve_interpolation_with_end_derivatives(points,
                                                               degree,
                                                               start_derivative,
                                                               end_derivative,
                                                               knot_style,
                                                               periodic=periodic)
    elif start_derivative is not None or end_derivative is not None:
        raise ValueError("Please pass start- AND end derivatives")
    else:
        return global_curve_interpolation(points, degree, knot_style=knot_style, periodic=periodic)


if __name__ == "__main__":
    from compas_nurbs import Curve

    knot_style = CurveKnotStyle.Uniform
    points = [[0.0, 0.0, 0.0], [0.973412, 1.962979, 0.0], [-0.66136, 2.434672, 0.0], [-2.34574, 1.016886, 0.0], [-3.513204, -0.855328, 0.0], [-4.0, -3.0, 0.0]]
    degree = 3

    control_points, knot_vector = global_curve_interpolation(points, degree, CurveKnotStyle.Uniform)
    crv = Curve(control_points, degree, knot_vector)

    start_derivative, end_derivative = [17.75, 10.79, 0.0], [-0.71, -12.62, 0.0]
    result = global_curve_interpolation_with_end_derivatives(points,
                                                             degree,
                                                             start_derivative,
                                                             end_derivative,
                                                             knot_style)

    crv = Curve(result[0], degree, result[1])
    D0, D1 = crv.derivatives_at([0, 1], order=1)[0]
    assert(allclose(D0, start_derivative))
    assert(allclose(D1, end_derivative))

    result = interpolate_curve(points, degree, knot_style)
    crv = Curve(result[0], degree, result[1])
    D0, D1 = crv.derivatives_at([0, 1], order=1)[0]
    assert(allclose(D0, start_derivative, tol=0.01))
    assert(allclose(D1, end_derivative, tol=0.01))



from compas_nurbs import Curve
from compas_nurbs.rhino import CurveArtist

control_points = [(0, 0, 0), (3, 4, 0), (-1, 4, 0),
                  (-4, 0, 0), (-4, -3, 0)]
curve = Curve(control_points, 3)

artist = CurveArtist(curve)
artist.draw()
artist.draw_points()
artist.redraw()




from compas.geometry import Vector
from compas.geometry import Circle
from compas.geometry import Plane


class CurveCurvature(object):
    """A container class with several curve curvature quantities.
    """

    def __init__(self, curvature, frame):
        self.curvature = curvature
        self.frame = frame

    @property
    def tangent(self):
        return self.frame.xaxis

    @property
    def normal(self):
        return self.frame.yaxis

    @property
    def binormal(self):
        return self.frame.zaxis

    @property
    def radius(self):
        return 1./self.curvature

    @property
    def osculating_circle(self):
        center = self.frame.point + self.radius * self.normal
        return Circle(Plane(center, self.frame.zaxis), self.radius)


class SurfaceCurvature(object):
    """A container class with several surface curvature quantities.
    """

    def __init__(self, kappa, direction, normal, mean, gauss):
        self.normal = Vector(*normal)
        self.kappa = list(kappa)
        self.direction = [Vector(*d) for d in direction]
        self.mean = mean
        self.gauss = gauss

    @property
    def osculating_circle(self):
        raise NotImplementedError



# TODO: move those functions to compas
from functools import reduce
from operator import mul


def linspace(start, stop, num, endpoint=True):
    """ Returns a list of evenly spaced numbers over a specified interval.

    Inspired from Numpy's linspace function: https://github.com/numpy/numpy/blob/master/numpy/core/function_base.py

    Parameters
    ----------
    start : float
        The starting value of the sequence.
    stop : float
        The end value of the sequence.
    num : int
        Number of samples to generate.
    endpoint : bool, optional
        If True, `stop` is the last sample. Otherwise, it is not included.
        Default is True.

    Returns
    -------
    samples : list
        There are `num` equally spaced samples in the closed interval
        ``[start, stop]`` or the half-open interval ``[start, stop)``
        (depending on whether `endpoint` is True or False).

    Examples
    --------
    >>> linspace(2.0, 3.0, 5)
    [2.0, 2.25, 2.5, 2.75, 3.0]
    >>> linspace(2.0, 3.0, 5, endpoint=False)
    [2.0, 2.2, 2.4, 2.6, 2.8]
    """
    if num < 0:
        raise ValueError("Number of samples, %s, must be non-negative." % num)
    start = float(start)
    stop = float(stop)
    if abs(start - stop) <= 10e-8:
        return [start]
    num = int(num)
    if num > 1:
        div = (num - 1) if endpoint else num
        delta = stop - start
        return [start + (float(x) * float(delta) / float(div)) for x in range(num)]
    else:
        return [start]


def unflatten(array, n):
    if len(array) % n:
        raise ValueError("The length of the array must be a factor of n: %d %% %d == 0" % (len(array), n))
    return [array[i:i + n] for i in range(0, len(array), n)]


def prod(array):
    return reduce(lambda x, y: x * y, array)


def reshape(lst, shape):
    """

    Examples
    --------
    >>> a = [1, 2, 3, 4, 5, 6]
    >>> reshape(a, (2, 3))
    [[1, 2, 3], [4, 5, 6]]
    >>> reshape(a, (3, 2))
    [[1, 2], [3, 4], [5, 6]]
    """
    if len(shape) == 1:
        return lst
    if len(lst) != prod(shape):
        raise ValueError("ValueError: cannot reshape array of size %d into shape %s" % (len(lst), shape))
    n = reduce(mul, shape[1:])
    return [reshape(lst[i * n:(i + 1) * n], shape[1:]) for i in range(len(lst) // n)]


if __name__ == '__main__':
    import doctest
    doctest.testmod(globs=globals())
